package entity

/*
ServiceError defines how internal service errors look
*/
type ServiceError struct {
	errorMessage string
	code         ErrorCode
}

/*
ErrorCode is an error code
*/
type ErrorCode int

const (
	// TranslationError is thrown if there is an error translating
	TranslationError ErrorCode = iota
	// MalformedRequestError is thrown if the query request is malformed
	MalformedRequestError
	// DatabaseCredentialsError is thrown if the database credentials can't be retrieved
	DatabaseCredentialsError
	// DatabaseExecutionError is thrown if there is an error while executing the query on the database
	DatabaseExecutionError
	// MessageHeaderError is thrown if headers are missing from request
	MessageHeaderError
)

/*
NewError creates a new ServiceError
	errorMessage: string, the error message
	code: int, the error code
	Return: error, a created error
*/
func NewError(errorMessage string, code ErrorCode) error {
	return &ServiceError{
		errorMessage: errorMessage,
		code:         TranslationError,
	}
}

func (se *ServiceError) Error() string {
	return se.errorMessage
}
